//    Copyright (C) 2009  Regis COSNIER
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using OpenTK.Graphics;
using OpenTK.Math;

namespace StreetViewMapper
{
	public class Object3D : ICloneable
	{
		public enum FaceType { Quads, Triangles, Lines, Points };
		public FaceType faceType = FaceType.Triangles;

		public string name;
		public float[] vertex = new float[0];
		public float[] normal = new float[0];
		public float[] texture = new float[0];
		public int[] indices = new int[0];
		public bool showObject = true;
		public Material material = new Material();

		public object Clone()
		{
			Object3D clone = new Object3D();
			clone.faceType = this.faceType;
			clone.name = (string)this.name.Clone();
			clone.vertex = (float[])this.vertex.Clone();
			clone.normal = (float[])this.normal.Clone();
			clone.texture = (float[])this.texture.Clone();
			clone.indices = (int[])this.indices.Clone();
			return clone;
		}

		public void Display()
		{
			Display(false);
		}

		public void Display(bool noMaterial)
		{
			if (!noMaterial)
				material.Begin();

			GL.EnableClientState(EnableCap.VertexArray);
			GL.VertexPointer(3,  VertexPointerType.Float, 0, vertex);

			if (normal.Length > 0)
			{
				GL.EnableClientState(EnableCap.NormalArray);
				GL.NormalPointer(NormalPointerType.Float, 0, normal);
			}

			if (texture.Length > 0)
			{
				GL.EnableClientState(EnableCap.TextureCoordArray);
				GL.TexCoordPointer(2, TexCoordPointerType.Float, 0, texture);
			}

			BeginMode elementType = 0;
			switch (faceType)
			{
				case FaceType.Quads: elementType = BeginMode.Quads; break;
                case FaceType.Triangles: elementType = BeginMode.Triangles; break;
                case FaceType.Lines: elementType = BeginMode.Lines; break;
				case FaceType.Points: elementType = BeginMode.Points; break;
			}
			GL.DrawElements(elementType, indices.Length, DrawElementsType.UnsignedInt, indices);

			GL.DisableClientState(EnableCap.VertexArray);

			if (normal.Length > 0)
				GL.DisableClientState(EnableCap.NormalArray);

			if (texture.Length > 0)
				GL.DisableClientState(EnableCap.TextureCoordArray);

			if(!noMaterial)
				material.End();
		}

		public void DisplayWireframe()
		{
			GL.EnableClientState(EnableCap.VertexArray);
			GL.VertexPointer(3, VertexPointerType.Float, 0, vertex);

			BeginMode elementType = BeginMode.Lines;

			GL.DrawElements(elementType, indices.Length, DrawElementsType.UnsignedInt, indices);

			GL.DisableClientState(EnableCap.VertexArray);
		}

		public void ComputeFlatNormals()
		{
			normal = new float[vertex.Length];

			switch (this.faceType)
			{
				case FaceType.Triangles:
					for (int i = 0; i < indices.Length; i += 3)
					{
						int index0 = 3 * indices[i + 0];
						int index1 = 3 * indices[i + 1];
						int index2 = 3 * indices[i + 2];

						Vector3 e = new Vector3(vertex[index0], vertex[index0 + 1], vertex[index0 + 2]);
						Vector3 e1 = new Vector3(vertex[index1], vertex[index1 + 1], vertex[index1 + 2]) - e;
						Vector3 e2 = new Vector3(vertex[index2], vertex[index2 + 1], vertex[index2 + 2]) - e;

						Vector3 n = Vector3.Cross(e1, e2);
						float length = n.Length;
						if (length > 0.0f)
						{
							n /= length;

							normal[index0] = n.X;
							normal[index0 + 1] = n.Y;
							normal[index0 + 2] = n.Z;
							normal[index1] = n.X;
							normal[index1 + 1] = n.Y;
							normal[index1 + 2] = n.Z;
							normal[index2] = n.X;
							normal[index2 + 1] = n.Y;
							normal[index2 + 2] = n.Z;
						}
					}
					break;
				case FaceType.Quads:
					for (int i = 0; i < indices.Length; i += 4)
					{
						int index0 = 3 * indices[i + 0];
						int index1 = 3 * indices[i + 1];
						int index2 = 3 * indices[i + 2];
						int index3 = 3 * indices[i + 3];

						Vector3 e0 = new Vector3(vertex[index0], vertex[index0 + 1], vertex[index0 + 2]);
						Vector3 e1 = new Vector3(vertex[index1], vertex[index1 + 1], vertex[index1 + 2]) - e0;
						Vector3 e2 = new Vector3(vertex[index2], vertex[index2 + 1], vertex[index2 + 2]) - e0;
						Vector3 e3 = new Vector3(vertex[index3], vertex[index3 + 1], vertex[index3 + 2]) - e0;

						Vector3 n = Vector3.Cross(e1, e2);
						float length = n.Length;
						if (length > 0.0f)
						{
							n /= length;

							normal[index0] = n.X;
							normal[index0 + 1] = n.Y;
							normal[index0 + 2] = n.Z;
							normal[index1] = n.X;
							normal[index1 + 1] = n.Y;
							normal[index1 + 2] = n.Z;
							normal[index2] = n.X;
							normal[index2 + 1] = n.Y;
							normal[index2 + 2] = n.Z;
						}

						n = Vector3.Cross(e2, e3);
						length = n.Length;
						if (length > 0.0f)
						{
							n /= length;

							normal[index0] = n.X;
							normal[index0 + 1] = n.Y;
							normal[index0 + 2] = n.Z;
							normal[index2] = n.X;
							normal[index2 + 1] = n.Y;
							normal[index2 + 2] = n.Z;
							normal[index3] = n.X;
							normal[index3 + 1] = n.Y;
							normal[index3 + 2] = n.Z;
						}
					}
					break;
			}
		}

		public void ComputeSmoothNormals()
		{
			normal = new float[vertex.Length];

			switch (this.faceType)
			{
				case FaceType.Triangles:
					for (int i = 0; i < indices.Length; i += 3)
					{
						int index0 = 3 * indices[i + 0];
						int index1 = 3 * indices[i + 1];
						int index2 = 3 * indices[i + 2];

						Vector3 e = new Vector3(vertex[index0], vertex[index0 + 1], vertex[index0 + 2]);
						Vector3 e1 = new Vector3(vertex[index1], vertex[index1 + 1], vertex[index1 + 2]) - e;
						Vector3 e2 = new Vector3(vertex[index2], vertex[index2 + 1], vertex[index2 + 2]) - e;

						Vector3 n = Vector3.Cross(e1, e2);
						float length = n.Length;
						if (length > 0.0f)
						{
							n /= length;

							normal[index0] += n.X;
							normal[index0 + 1] += n.Y;
							normal[index0 + 2] += n.Z;
							normal[index1] += n.X;
							normal[index1 + 1] += n.Y;
							normal[index1 + 2] += n.Z;
							normal[index2] += n.X;
							normal[index2 + 1] += n.Y;
							normal[index2 + 2] += n.Z;
						}
					}

					for (int i = 0; i < normal.Length; i += 3)
					{
						Vector3 n = new Vector3(normal[i], normal[i + 1], normal[i + 2]);
						float length = n.Length;
						if (length > 0.0f) n /= length;
						normal[i] = n.X;
						normal[i + 1] = n.Y;
						normal[i + 2] = n.Z;
					}
					break;
				case FaceType.Quads:
					for (int i = 0; i < indices.Length; i += 4)
					{
						int index0 = 3 * indices[i + 0];
						int index1 = 3 * indices[i + 1];
						int index2 = 3 * indices[i + 2];
						int index3 = 3 * indices[i + 3];

						Vector3 e0 = new Vector3(vertex[index0], vertex[index0 + 1], vertex[index0 + 2]);
						Vector3 e1 = new Vector3(vertex[index1], vertex[index1 + 1], vertex[index1 + 2]) - e0;
						Vector3 e2 = new Vector3(vertex[index2], vertex[index2 + 1], vertex[index2 + 2]) - e0;
						Vector3 e3 = new Vector3(vertex[index3], vertex[index3 + 1], vertex[index3 + 2]) - e0;

						Vector3 n = Vector3.Cross(e1, e2);
						float length = n.Length;
						if (length > 0.0f)
						{
							n /= length;

							normal[index0] += n.X;
							normal[index0 + 1] += n.Y;
							normal[index0 + 2] += n.Z;
							normal[index1] += n.X;
							normal[index1 + 1] += n.Y;
							normal[index1 + 2] += n.Z;
							normal[index2] += n.X;
							normal[index2 + 1] += n.Y;
							normal[index2 + 2] += n.Z;
						}

						n = Vector3.Cross(e2, e3);
						length = n.Length;
						if (length > 0.0f)
						{
							n /= length;

							normal[index0] += n.X;
							normal[index0 + 1] += n.Y;
							normal[index0 + 2] += n.Z;
							normal[index2] += n.X;
							normal[index2 + 1] += n.Y;
							normal[index2 + 2] += n.Z;
							normal[index3] += n.X;
							normal[index3 + 1] += n.Y;
							normal[index3 + 2] += n.Z;
						}
					}

					for (int i = 0; i < normal.Length; i += 3)
					{
						Vector3 n = new Vector3(normal[i], normal[i + 1], normal[i + 2]);
						float length = n.Length;
						if (length > 0.0f) n /= length;
						normal[i] = n.X;
						normal[i + 1] = n.Y;
						normal[i + 2] = n.Z;
					}
					break;
			}
		}

		//// The source code comes from Eric Lengyel at http://www.terathon.com/code/tangent.html
		//void Model::Geometry::ComputeTangents()
		//{
		//    Vector3 *tan1 = new Vector3[nNumVertices * 2];
		//    Vector3 *tan2 = tan1 + nNumVertices;
		//    memset(tan1, 0, nNumVertices * sizeof(Vector3) * 2);

		//    int numOfIndices = 3 * nNumFaces;
		//    for(int nIndice = 0; nIndice < numOfIndices; nIndice += 3)
		//    {
		//        long i1 = indices[nIndice + 0];
		//        long i2 = indices[nIndice + 1];
		//        long i3 = indices[nIndice + 2];

		//        const Point3& v1 = vertex[i1];
		//        const Point3& v2 = vertex[i2];
		//        const Point3& v3 = vertex[i3];

		//        const Point2& w1 = aTextures[i1];
		//        const Point2& w2 = aTextures[i2];
		//        const Point2& w3 = aTextures[i3];

		//        float x1 = v2.x - v1.x;
		//        float x2 = v3.x - v1.x;
		//        float y1 = v2.y - v1.y;
		//        float y2 = v3.y - v1.y;
		//        float z1 = v2.z - v1.z;
		//        float z2 = v3.z - v1.z;

		//        float s1 = w2.x - w1.x;
		//        float s2 = w3.x - w1.x;
		//        float t1 = w2.y - w1.y;
		//        float t2 = w3.y - w1.y;

		//        float r = 1.0F / (s1 * t2 - s2 * t1);
		//        Vector3 sdir((t2 * x1 - t1 * x2) * r, (t2 * y1 - t1 * y2) * r,
		//                (t2 * z1 - t1 * z2) * r);
		//        Vector3 tdir((s1 * x2 - s2 * x1) * r, (s1 * y2 - s2 * y1) * r,
		//                (s1 * z2 - s2 * z1) * r);

		//        tan1[i1] += sdir;
		//        tan1[i2] += sdir;
		//        tan1[i3] += sdir;

		//        tan2[i1] += tdir;
		//        tan2[i2] += tdir;
		//        tan2[i3] += tdir;
		//    }

		//    long count = nNumVertices;
		//    for (long a = 0; a < count; a++)
		//    {
		//        const Vector3 & n = normal[a];
		//        const Vector3 & t = tan1[a];

		//        // Gram-Schmidt orthogonalize
		//        aTangents[a] = Vector4((t - n * (n * t)).Normalize());

		//        // Calculate handedness
		//        aTangents[a].w = (((n % t) * tan2[a]) < 0.0F) ? -1.0F : 1.0F;
		//    }

		//    delete[] tan1;
		//}

		private float ComputeTriangleArea(ref Vector3 v0, ref Vector3 v1, ref Vector3 v2)
		{
			return Vector3.Cross(v1 - v0, v2 - v0).Length / 2.0f;
		}

		public float ComputeLength()
		{
			float accumulatedLength = 0.0f;
			switch (this.faceType)
			{
				case FaceType.Lines:
					for (int i = 0; i < indices.Length; i += 2)
					{
						int i1 = 3 * indices[i];
						int i2 = 3 * indices[i + 1];
						Vector3 v1 = new Vector3(vertex[i1], vertex[i1 + 1], vertex[i1 + 2]);
						Vector3 v2 = new Vector3(vertex[i2], vertex[i2 + 1], vertex[i2 + 2]);
						accumulatedLength += (v2 - v1).Length;
					}
					break;
			}
			return accumulatedLength;
		}

		public float ComputeArea()
		{
			float accumulatedArea = 0.0f;
			switch (this.faceType)
			{
				case FaceType.Triangles:
					for (int i = 0; i < indices.Length; i += 3)
					{
						int i1 = 3 * indices[i];
						int i2 = 3 * indices[i + 1];
						int i3 = 3 * indices[i + 2];
						Vector3 v1 = new Vector3(vertex[i1], vertex[i1 + 1], vertex[i1 + 2]);
						Vector3 v2 = new Vector3(vertex[i2], vertex[i2 + 1], vertex[i2 + 2]);
						Vector3 v3 = new Vector3(vertex[i3], vertex[i3 + 1], vertex[i3 + 2]);
						accumulatedArea += Vector3.Cross(v2 - v1, v3 - v1).Length / 2.0f;
					}
					break;
				case FaceType.Quads:
					for (int i = 0; i < indices.Length; i += 4)
					{
						int i1 = 3 * indices[i];
						int i2 = 3 * indices[i + 1];
						int i3 = 3 * indices[i + 2];
						int i4 = 3 * indices[i + 3];
						Vector3 v1 = new Vector3(vertex[i1], vertex[i1 + 1], vertex[i1 + 2]);
						Vector3 v2 = new Vector3(vertex[i2], vertex[i2 + 1], vertex[i2 + 2]);
						Vector3 v3 = new Vector3(vertex[i3], vertex[i3 + 1], vertex[i3 + 2]);
						Vector3 v4 = new Vector3(vertex[i4], vertex[i4 + 1], vertex[i4 + 2]);
						accumulatedArea += Vector3.Cross(v2 - v1, v3 - v1).Length / 2.0f;
						accumulatedArea += Vector3.Cross(v3 - v1, v4 - v1).Length / 2.0f;
					}
					break;
			}
			return accumulatedArea;
		}

		// Author: Kyle Owen
		// From: www.gamedev.net - Area and Volume Calculations
		// http://www.gamedev.net/reference/articles/article2247.asp
		// This function uses Gauss's Theorem to calculate the volume of a body
		// enclosed by a set of triangles. The triangle set must form a closed
		// surface in R3 and be outward facing. Outward facing triangles index
		// their vertices in a counterclockwise order where the x-axis points
		// left, the y-axis point up and the z-axis points toward you (rhs).
		public float ComputeVolume()
		{
			float accumulatedVolume = 0.0f;
			switch (this.faceType)
			{
				case FaceType.Triangles:
					for (int i = 0; i < indices.Length; i += 3)
					{
						int i1 = 3 * indices[i];
						int i2 = 3 * indices[i + 1];
						int i3 = 3 * indices[i + 2];
						Vector3 v1 = new Vector3(vertex[i1], vertex[i1 + 1], vertex[i1 + 2]);
						Vector3 v2 = new Vector3(vertex[i2], vertex[i2 + 1], vertex[i2 + 2]);
						Vector3 v3 = new Vector3(vertex[i3], vertex[i3 + 1], vertex[i3 + 2]);

						accumulatedVolume += ((v2.Y - v1.Y) * (v3.Z - v1.Z) - (v2.Z - v1.Z) * (v3.Y - v1.Y)) * (v1.X + v2.X + v3.X);
					}
					break;
				case FaceType.Quads:
					for (int i = 0; i < indices.Length; i += 4)
					{
						int i1 = 3 * indices[i];
						int i2 = 3 * indices[i + 1];
						int i3 = 3 * indices[i + 2];
						int i4 = 3 * indices[i + 3];
						Vector3 v1 = new Vector3(vertex[i1], vertex[i1 + 1], vertex[i1 + 2]);
						Vector3 v2 = new Vector3(vertex[i2], vertex[i2 + 1], vertex[i2 + 2]);
						Vector3 v3 = new Vector3(vertex[i3], vertex[i3 + 1], vertex[i3 + 2]);
						Vector3 v4 = new Vector3(vertex[i4], vertex[i4 + 1], vertex[i4 + 2]);

						accumulatedVolume += ((v2.Y - v1.Y) * (v3.Z - v1.Z) - (v2.Z - v1.Z) * (v3.Y - v1.Y)) * (v1.X + v2.X + v3.X);
						accumulatedVolume += ((v3.Y - v1.Y) * (v4.Z - v1.Z) - (v3.Z - v1.Z) * (v4.Y - v1.Y)) * (v1.X + v3.X + v4.X);
					}
					break;
			}
			return accumulatedVolume / 6.0f;
		}
	}

	public class Mesh
	{
		public List<Object3D> objects = new List<Object3D>();
		public Dictionary<string, Material> materials = new Dictionary<string, Material>();

		public void Init(string filename)
		{
			Trace.WriteLine(string.Format("Loading mesh: {0}", filename));
			
			using (StreamReader streamReader = new StreamReader(filename))
			{
				string line;
				string[] tokens, subTokens;
				float x, y, z;
				Object3D o = null;
				List<float> vertices0 = null, normals0 = null, textureCoords0 = null;
				List<float> vertices = null, normals = null, textureCoords = null;
				List<int> indices = null;
				int vertexIndexOffset = 1, textureIndexOffset = 1, normalIndexOffset = 1;
				int currentVertexIndexOffset = 1, currentTextureIndexOffset = 1, currentNormalIndexOffset = 1;
				int fullVertexIndex = 0;
				int verticesCount = 0, textureCoordsCount = 0, normalsCount = 0;

				Dictionary<UniqueVertexIndex, int> mapIndexByIndices = new Dictionary<UniqueVertexIndex, int>();

				while (true)
				{
					line = streamReader.ReadLine();
					if (line == null || line[0] == 'o')
					{
						if (o != null)
						{
							//OK, this is the end of a mesh object

							Trace.WriteLine(string.Format("\t\tIn file  : {0} vertices, {1} texture coordinates, {2} normals", vertices0.Count / 3, textureCoords0.Count / 2, normals0.Count / 3));
							Trace.WriteLine(string.Format("\t\tUnindexed: {0} vertices, {1} texture coordinates, {2} normals", vertices.Count / 3, textureCoords.Count / 2, normals.Count / 3));

							if (indices.Count > 0)
							{
								o.vertex = vertices.ToArray();
								o.texture = textureCoords.ToArray();
								o.normal = normals.ToArray();
								o.indices = indices.ToArray();
							}
							else if (o.faceType == Object3D.FaceType.Points)
							{
								o.vertex = vertices0.ToArray();
								o.texture = textureCoords0.ToArray();
								o.normal = normals0.ToArray();
								o.indices = new int[vertices0.Count / 3];
								for (int i = 0; i < vertices0.Count; i += 3)
									o.indices[i] = (int)i;
							}
							objects.Add(o);

							vertexIndexOffset = currentVertexIndexOffset;
							textureIndexOffset = currentTextureIndexOffset;
							normalIndexOffset = currentNormalIndexOffset;

							verticesCount += o.vertex.Length / 3;
							textureCoordsCount += o.texture.Length / 2;
							normalsCount += o.normal.Length / 3;
						}

						if (line == null)
							break;

						//OK, this is a new mesh object

						o = new Object3D();
						tokens = line.Split(new char[] { ' ', '\t' });
						o.name = tokens[1];
						o.faceType = Object3D.FaceType.Points;

						Trace.WriteLine(string.Format("\tLoading: {0}", o.name));

						vertices0 = new List<float>();
						normals0 = new List<float>();
						textureCoords0 = new List<float>();

						vertices = new List<float>();
						normals = new List<float>();
						textureCoords = new List<float>();

						indices = new List<int>();
						fullVertexIndex = 0;

						mapIndexByIndices.Clear();
					}
					else if (line[0] == 'v' && line[1] == 't')
					{
						tokens = line.Split(new char[] { ' ', '\t' });

						// Add the mesh texture coordinates
						textureCoords0.Add(float.Parse(tokens[1]));
						textureCoords0.Add(float.Parse(tokens[2]));
						//textureCoords.Add(float.Parse(tokens[3]));

						currentTextureIndexOffset++;
					}
					else if (line[0] == 'v' && line[1] == 'n')
					{
						tokens = line.Split(new char[] { ' ', '\t' });

						// Add the mesh normal
						normals0.Add(float.Parse(tokens[1]));
						normals0.Add(float.Parse(tokens[2]));
						normals0.Add(float.Parse(tokens[3]));

						currentNormalIndexOffset++;
					}
					else if (line[0] == 'v')
					{
						tokens = line.Split(new char[] { ' ', '\t' });

						// Add the mesh vertex
						vertices0.Add(x = float.Parse(tokens[1]));
						vertices0.Add(y = float.Parse(tokens[2]));
						vertices0.Add(z = float.Parse(tokens[3]));

						currentVertexIndexOffset++;
					}
					else if (line[0] == 'f')
					{
						tokens = line.Split(new char[] { ' ', '\t' });
						int numberOfVertex = tokens.Length - 1;

						for (int i = 0; i < numberOfVertex; i++)
						{
							int textureIndex0 = 0;
							int normalIndex0 = 0;

							subTokens = tokens[i + 1].Split(new char[] { '/' });
							bool hasTextureCoordinates = (subTokens.Length > 1 && subTokens[1].Trim().Length > 0),
								hasNormal = (subTokens.Length > 2 && subTokens[2].Trim().Length > 0);
							int vertexIndex0 = int.Parse(subTokens[0]);
							if (hasTextureCoordinates) textureIndex0 = int.Parse(subTokens[1]);
							if (hasNormal) normalIndex0 = int.Parse(subTokens[2]);

							int v0 = vertexIndex0 - vertexIndexOffset;
							int t0 = textureIndex0 - textureIndexOffset;
							int n0 = normalIndex0 - normalIndexOffset;

							UniqueVertexIndex vi0 = new UniqueVertexIndex(v0, t0, n0,
								hasTextureCoordinates ? textureCoords0[2 * t0] : 0.0f,
								hasTextureCoordinates ? textureCoords0[2 * t0 + 1] : 0.0f);

							int i0 = v0;
							if (!mapIndexByIndices.TryGetValue(vi0, out i0))
							{
								vertices.Add(vertices0[(int)(v0 * 3)]);
								vertices.Add(vertices0[(int)(v0 * 3 + 1)]);
								vertices.Add(vertices0[(int)(v0 * 3 + 2)]);
								if (hasTextureCoordinates)
								{
									textureCoords.Add(textureCoords0[(int)(t0 * 2)]);
									textureCoords.Add(textureCoords0[(int)(t0 * 2 + 1)]);
								}
								if (hasNormal)
								{
									normals.Add(normals0[(int)(n0 * 3)]);
									normals.Add(normals0[(int)(n0 * 3 + 1)]);
									normals.Add(normals0[(int)(n0 * 3 + 2)]);
								}
								mapIndexByIndices[vi0] = i0 = fullVertexIndex++;
							}

							indices.Add(i0);
						}

						switch (tokens.Length)
						{
							case 3: o.faceType = Object3D.FaceType.Lines; break;
							case 4: o.faceType = Object3D.FaceType.Triangles; break;
							case 5: o.faceType = Object3D.FaceType.Quads; break;
						}
					}
					else if (line[0] == 'm' && line.StartsWith("mtllib", StringComparison.InvariantCultureIgnoreCase))
					{
						tokens = line.Split(new char[] { ' ', '\t' });
						if (tokens.Length > 1)
						{
							LoadMaterialLibrary(Path.GetDirectoryName(filename) + "\\" + tokens[1]);
						}
					}
					else if (line[0] == 'u' && line.StartsWith("usemtl", StringComparison.InvariantCultureIgnoreCase))
					{
						tokens = line.Split(new char[] { ' ', '\t' });
						if (tokens.Length > 1)
						{
							if (!materials.TryGetValue(tokens[1], out o.material))
								o.material = new Material();
						}
					}
				}

				Trace.WriteLine(string.Format("\tCumulated  : {0} vertices, {1} texture coordinates, {2} normals", verticesCount, textureCoordsCount, normalsCount));
			}
		}

		struct UniqueVertexIndex
		{
			public UniqueVertexIndex(int vertex, int texture, int normal, float tu, float tv)
			{
				this.vertex = vertex;
				//this.texture = texture;
				this.normal = normal;
				this.tu = tu;
				this.tv = tv;
			}
			public int vertex;
			//public int texture;
			public int normal;
			public float tu;
			public float tv;
		};

		private void LoadMaterialLibrary(string filename)
		{
			Trace.WriteLine(string.Format("\tLoading mesh material: {0}", filename));

			using (StreamReader streamReader = new StreamReader(filename))
			{
				string line;
				string[] tokens;
				Material material = null;

				Dictionary<UniqueVertexIndex, int> mapIndexByIndices = new Dictionary<UniqueVertexIndex, int>();

				while (true)
				{
					line = streamReader.ReadLine();
					if (line == null || line.StartsWith("newmtl", StringComparison.InvariantCultureIgnoreCase))
					{
						if (material != null)
						{
							//OK, this is the end of a material

							switch (material.materialType)
							{
								case Material.MaterialType.Diffuse:
									Trace.WriteLine(string.Format("\t\t\tDiffuse  : {0}, {1}, {2}", material.diffuseColor.X, material.diffuseColor.Y, material.diffuseColor.Z));
									break;
								case Material.MaterialType.Texture:
									Trace.WriteLine(string.Format("\t\t\tTexture  : {0} ({1})", material.name , material.textureFileName));
									break;
							}

							material.Init();
							materials.Add(material.name, material);
						}

						if (line == null)
							break;

						//OK, this is a new material

						material = new Material();
						tokens = line.Split(new char[] { ' ', '\t' });
						material.name = tokens[1];

						Trace.WriteLine(string.Format("\t\tLoading material: {0}", material.name));
					}
					else if (line.StartsWith("Kd", StringComparison.InvariantCultureIgnoreCase))
					{
						tokens = line.Split(new char[] { ' ', '\t' });

						material.diffuseColor = new Vector4(float.Parse(tokens[1]), float.Parse(tokens[2]), float.Parse(tokens[3]), 1.0f);
					}
					else if (line.StartsWith("map_Kd", StringComparison.InvariantCultureIgnoreCase))
					{
						tokens = line.Split(new char[] { ' ', '\t' });

						material.textureFileName = Path.GetDirectoryName(filename) + "\\" + tokens[1];
						material.materialType = Material.MaterialType.Texture;
					}
				}
			}
		}

		public void Release()
		{
			objects.Clear();
			materials.Clear();
		}

		public void Display()
		{
			Display(false, false);
		}

		public void Display(bool onlyFaces, bool forceShowingAllObject)
		{
			if (onlyFaces)
			{
				foreach (Object3D o in objects)
					if ((o.faceType == Object3D.FaceType.Triangles || o.faceType == Object3D.FaceType.Quads) &&
						(forceShowingAllObject || o.showObject))
						o.Display();
			}
			else
			{
				foreach (Object3D o in objects)
					if (forceShowingAllObject || o.showObject)
						o.Display();
			}
		}
	}
}
