//    Copyright (C) 2009  Regis COSNIER
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

using System;
using System.Drawing;
using OpenTK.Graphics;
using OpenTK.Math;

namespace StreetViewMapper
{
	public class RTTexture
	{
		Vector4 clearColor;

		int[] saveViewport = new int[4];
		int[] oldViewport = new int[] { 0, 0, 0, 0 };
		int mapWidth, mapHeight, textureName;

		bool useFBO = false;
		int framebufferID;			// FBO name
		int renderbufferID;			// renderbuffer object name
		int maxTexSize;

		public RTTexture()
		{
			mapWidth = mapHeight = 512;
			clearColor = new Vector4(0.0f, 0.0f, 0.0f, 1.0f);

			Square = false;
		}

		public int Width { get { return mapWidth; } }
		public int Height { get { return mapHeight; } }
		public bool Square { get; set; }


		public void Init()
		{
			Init(-1, -1);
		}

		public void Init(int width, int height)
		{
			Release();

			useFBO = GLHelpers.IsExtensionSupported("GL_EXT_framebuffer_object");

			int[] cId = new int[1];
			GL.GetInteger(GetPName.MaxTextureSize, cId);
			maxTexSize = cId[0];
            if (useFBO)
            {
                // maximum allowed size for FBO renderbuffer
				//GL.GetInteger(GetPName.MaxRenderbufferSizeExt, cId);
				GL.GetInteger(GetPName.MaxRenderbufferSize, cId);
				maxTexSize = Math.Min(maxTexSize, cId[0]);
            }
            // performance suffers too much at higher texture sizes
            maxTexSize = Math.Min(maxTexSize, 2048);

			if (width > 0)
				mapWidth = width;
			else
				mapWidth = maxTexSize;
			if (height > 0)
				mapHeight = height;
			else
				mapHeight = maxTexSize;


			// Create a texture object for the reflection texture:
			GL.GenTextures(1, cId);
			textureName = cId[0];
			GL.BindTexture(TextureTarget.Texture2D, textureName);

			GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureWrapS, (int)TextureWrapMode.ClampToEdge);
			GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureWrapT, (int)TextureWrapMode.ClampToEdge);
			GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureMagFilter, (int)TextureMinFilter.Linear);
			GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureMinFilter, (int)TextureMinFilter.Linear);

			GL.CopyTexImage2D(TextureTarget.Texture2D, 0, PixelInternalFormat.Rgba, 0, 0, mapWidth, mapHeight, 0);
			GL.BindTexture(TextureTarget.Texture2D, 0);

			// Set up some renderbuffer state
			if (useFBO)
			{
				GL.Ext.GenFramebuffers(1, cId);
				framebufferID = cId[0];
				GL.Ext.BindFramebuffer(FramebufferTarget.FramebufferExt, framebufferID);
				GL.Ext.GenRenderbuffers(1, cId);
				renderbufferID = cId[0];
				GL.Ext.BindRenderbuffer(RenderbufferTarget.RenderbufferExt, renderbufferID);
				GL.Ext.RenderbufferStorage(RenderbufferTarget.RenderbufferExt, RenderbufferStorage.DepthComponent24, mapWidth, mapHeight);
				GL.Ext.FramebufferRenderbuffer(FramebufferTarget.FramebufferExt, FramebufferAttachment.DepthAttachmentExt, RenderbufferTarget.RenderbufferExt, renderbufferID);
				GL.Ext.FramebufferTexture2D(FramebufferTarget.FramebufferExt, FramebufferAttachment.ColorAttachment0Ext, TextureTarget.Texture2D, textureName, 0);
				FramebufferErrorCode fboStatus = GL.Ext.CheckFramebufferStatus(FramebufferTarget.FramebufferExt);
				if (fboStatus !=  FramebufferErrorCode.FramebufferCompleteExt)
				{
					//fprintf(stderr, "FBO Error!\n");
					useFBO = false;
				}
				GL.Ext.BindFramebuffer(FramebufferTarget.FramebufferExt, 0);
			}
		}

		public void Release()
		{
			if (framebufferID > 0)
				GL.Ext.DeleteFramebuffers(1, new int[] { framebufferID });
			if (renderbufferID > 0)
				GL.Ext.DeleteRenderbuffers(1, new int[] { renderbufferID });

			if (textureName > 0)
			{
				GL.DeleteTextures(1, new int[] { this.textureName });
				this.textureName = 0;
			}
		}

		public void BeginDraw()
		{
			GL.GetInteger(GetPName.Viewport, saveViewport);
			if (oldViewport[2] != saveViewport[2] || oldViewport[3] != saveViewport[3])
			{
				saveViewport.CopyTo(oldViewport, 0);

				if (useFBO)
				{
					// shadow map is limited by max supported renderbuffer size
					//mapWidth = mapHeight = maxTexSize;
				}
				else
				{
					// shadow map is limited to size of window
					mapWidth = saveViewport[2];
					mapHeight = saveViewport[3];

					//if (!npotTexturesAvailable)
					{
						// Find the largest power of two that will fit in window.
						// Try each width until we get one that's too big
						int i = 0;
						while ((1 << i) <= mapWidth) i++;
						mapWidth = (1 << (i - 1));

						// Now for height
						i = 0;
						while ((1 << i) <= mapHeight) i++;
						mapHeight = (1 << (i - 1));
					}

					if (Square)
						mapHeight = mapWidth = Math.Min(mapWidth, mapHeight);

					if (mapWidth > maxTexSize) mapWidth = maxTexSize;
					if (mapHeight > maxTexSize) mapHeight = maxTexSize;
				}

				GL.BindTexture(TextureTarget.Texture2D, textureName);
				GL.CopyTexImage2D(TextureTarget.Texture2D, 0, PixelInternalFormat.Rgba, 0, 0, mapWidth, mapHeight, 0);
				GL.BindTexture(TextureTarget.Texture2D, 0);

				if (useFBO)
				{
					GL.Ext.BindRenderbuffer(RenderbufferTarget.RenderbufferExt, renderbufferID);
					GL.Ext.RenderbufferStorage(RenderbufferTarget.RenderbufferExt, RenderbufferStorage.DepthComponent24, mapWidth, mapHeight);
					FramebufferErrorCode fboStatus = GL.Ext.CheckFramebufferStatus(FramebufferTarget.FramebufferExt);
					if (fboStatus != FramebufferErrorCode.FramebufferCompleteExt)
					{
						//fprintf(stderr, "FBO Error!\n");
						useFBO = false;
					}
					GL.Ext.BindFramebuffer(FramebufferTarget.FramebufferExt, 0);
				}
			}
			GL.Viewport(0, 0, mapWidth, mapHeight);

			if (useFBO)
				GL.Ext.BindFramebuffer(FramebufferTarget.FramebufferExt, framebufferID);

			float[] arSaveClearColor = new float[4];
			GL.GetFloat(GetPName.ColorClearValue, arSaveClearColor);
			GL.ClearColor(clearColor.X, clearColor.Y, clearColor.Z, clearColor.W);
			GL.Clear(ClearBufferMask.ColorBufferBit | ClearBufferMask.DepthBufferBit);
			GL.ClearColor(arSaveClearColor[0], arSaveClearColor[1], arSaveClearColor[2], arSaveClearColor[3]);
		}

		public Bitmap GrabScreenshot()
		{
			Bitmap bmp = new Bitmap(mapWidth, mapHeight);
			System.Drawing.Imaging.BitmapData data = bmp.LockBits(new Rectangle(0, 0, mapWidth, mapHeight),
																	System.Drawing.Imaging.ImageLockMode.WriteOnly,
																	System.Drawing.Imaging.PixelFormat.Format24bppRgb);
			GL.ReadPixels(0, 0, mapWidth, mapHeight, PixelFormat.Bgr, PixelType.UnsignedByte, data.Scan0);
			bmp.UnlockBits(data);
			bmp.RotateFlip(RotateFlipType.RotateNoneFlipY);
			return bmp;
		}

		public void EndDraw()
		{
			// copy buffer into texture
			GL.BindTexture(TextureTarget.Texture2D, textureName);
			GL.CopyTexImage2D(TextureTarget.Texture2D, 0, PixelInternalFormat.Rgb, 0, 0, mapWidth, mapHeight, 0);
			GL.BindTexture(TextureTarget.Texture2D, 0);

			if (useFBO)
				GL.Ext.BindFramebuffer(FramebufferTarget.FramebufferExt, 0);

			GL.Viewport(saveViewport[0], saveViewport[1], saveViewport[2], saveViewport[3]);
		}

		public void Begin()
		{
			GL.Enable(EnableCap.Texture2D);
			GL.BindTexture(TextureTarget.Texture2D, textureName);
			//GL.TexEnvi(TEXTURE_ENV, TEXTURE_ENV_MODE, MODULATE);
		}

		public void End()
		{
			GL.BindTexture(TextureTarget.Texture2D, 0);
			GL.Disable(EnableCap.Texture2D);
		}
	}
}
