//    Copyright (C) 2009  Regis COSNIER
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

using System.Diagnostics;
using System.IO;
using System.Text;
using OpenTK.Graphics;

namespace StreetViewMapper
{
	public class Shader
	{
		bool isLoaded;

		int shaderId;
		int vertexShaderId;
		int fragmentShaderId;

		public Shader()
		{
			shaderId = vertexShaderId = fragmentShaderId = 0;
			isLoaded = false;
		}

		public bool IsLoaded { get { return isLoaded; } }

		enum ShaderObjectType : int
		{
			VertexShader = ((int)0x8b31),
			FragmentShader = ((int)0x8b30),
			GeometryShader = ExtGeometryShader4.GeometryShaderExt,
		}

		public void LoadShaderFiles(string vertexShader, string fragmentShader)
		{
			Trace.Write(string.Format("\tLoading the GLSL vertex shader ({0})\n", vertexShader));

			string loadedSourceCode = LoadCodeSource(vertexShader);
			if (loadedSourceCode.Length == 0)
			{
				Trace.Write(string.Format("\tCannot load the GLSL vertex shader ({0})\n", vertexShader));
				Release();
				return;
			}

			vertexShader = loadedSourceCode;

			Trace.Write(string.Format("\tLoading the GLSL fragment shader ({0})\n", fragmentShader));
			loadedSourceCode = LoadCodeSource(fragmentShader);
			if (loadedSourceCode.Length == 0)
			{
				Trace.Write(string.Format("\tCannot load the GLSL fragment shader ({0})\n", fragmentShader));
				Release();
				return;
			}

			fragmentShader = loadedSourceCode;

			Load(vertexShader, fragmentShader);
		}

		public void Load(string vertexShader, string fragmentShader)
		{
			Release();

			Trace.Write(string.Format("Loading and compiling GLSL vertex and fragment shader\n"));
			if (GLHelpers.IsExtensionSupported("GL_ARB_shader_objects") &&
				GLHelpers.IsExtensionSupported("GL_ARB_vertex_shader") &&
				GLHelpers.IsExtensionSupported("GL_ARB_fragment_shader") &&
				GLHelpers.IsExtensionSupported("GL_ARB_shading_language_100"))
			{
				vertexShaderId = GL.Arb.CreateShaderObject((ArbShaderObjects)ShaderObjectType.VertexShader);

				//UTF8Encoding utf8Encoding = new UTF8Encoding();
				//utf8Encoding.GetChars(

				string sourceCode = vertexShader;
				int[] lengths = new int[] { sourceCode.Length - 1 };
				//int length = 0;
				GL.Arb.ShaderSource(vertexShaderId, 1, new string[] { sourceCode }, lengths);
				GL.CompileShader(vertexShaderId);
				int[] result = new int[1] { -1 };
				GL.Arb.GetObjectParameter(vertexShaderId, ArbShaderObjects.ObjectCompileStatusArb, result);
				if (result[0] == 0)
				{
					TraceInfoLog(vertexShaderId);
					Release();
					return;
				}

				fragmentShaderId = GL.Arb.CreateShaderObject((ArbShaderObjects)ShaderObjectType.FragmentShader);

				sourceCode = fragmentShader;
				lengths = new int[] { sourceCode.Length - 1 };
				//length = 0;
				GL.Arb.ShaderSource(fragmentShaderId, 1, new string[] { sourceCode }, lengths);
				GL.Arb.CompileShader(fragmentShaderId);
				GL.Arb.GetObjectParameter(fragmentShaderId, ArbShaderObjects.ObjectCompileStatusArb, result);
				if (result[0] == 0)
				{
					TraceInfoLog(fragmentShaderId);
					Release();
					return;
				}

				// Create a program object and attach the two compiled shaders

				shaderId = GL.Arb.CreateProgramObject();
				GL.Arb.AttachObject(shaderId, vertexShaderId);
				GL.Arb.AttachObject(shaderId, fragmentShaderId);

				// Link the program object and print out the info log

				GL.Arb.LinkProgram(shaderId);
				GL.Arb.GetObjectParameter(shaderId, ArbShaderObjects.ObjectLinkStatusArb, result);
				TraceInfoLog(shaderId);

				if (result[0] == 0)
				{
					Release();
					return;
				}
				isLoaded = true;
			}
			else
				Trace.Write(string.Format("\tRequired ARB_shader_objects, ARB_vertex_shader, ARB_fragment_shader and ARB_shading_language_100 are not supported by your hardware.\n"));
		}

		public void Release()
		{
			if (!isLoaded) return;

			if (shaderId != 0)
			{
				if (vertexShaderId != 0)
				{
					GL.Arb.DetachObject(shaderId, vertexShaderId);
					GL.Arb.DeleteObject(vertexShaderId);
				}

				if (fragmentShaderId != 0)
				{
					GL.Arb.DetachObject(shaderId, fragmentShaderId);
					GL.Arb.DeleteObject(fragmentShaderId);
				}
				GL.Arb.DeleteObject(shaderId);
			}

			shaderId = vertexShaderId = fragmentShaderId = 0;
			isLoaded = false;
		}

		public void Begin()
		{
			if (!isLoaded) return;

			GL.Arb.UseProgramObject(shaderId);
		}
		public void End()
		{
			if (!isLoaded) return;

			GL.Arb.UseProgramObject(0);
		}


		static string LoadCodeSource(string filename)
		{
			string sourceCode = "";
			using (StreamReader streamReader = new StreamReader(filename))
			{
				sourceCode = streamReader.ReadToEnd();
			}
			return sourceCode;
		}

		public int GetUniformLocation(string uniformName)
		{
			int nLoc = GL.GetUniformLocation(shaderId, uniformName);
			if (nLoc == -1)
				Trace.Write(string.Format("No such uniform named \"{0}\"\n", uniformName));
			return nLoc;
		}

		public int GetAttribLocation(string attributeName)
		{
			int nLoc = GL.GetAttribLocation(shaderId, attributeName);
			if (nLoc == -1)
				Trace.Write(string.Format("No such attribute named \"{0}\"\n", attributeName));
			return nLoc;
		}

		// Print out the information log for a shader object or a program object
		void TraceInfoLog(int obj)
		{
			int[] result = new int[1];
			GL.Arb.GetObjectParameter(obj, ArbShaderObjects.ObjectInfoLogLengthArb, result);
			int infologLength = result[0];
			if (infologLength > 1) //0)
			{
			    //char[] infoLog = new char[infologLength];
				StringBuilder infoLog = new StringBuilder(infologLength);
				GL.Arb.GetInfoLog(obj, infologLength, result, infoLog);
				int charsWritten = result[0];
				Trace.Write(string.Format("{0}\n", infoLog));
			}
		}
	}
}
